#include <stdio.h>
#include "cuda.h"
#include "cuda_runtime_api.h"



__global__ void relax_on_CUDA(double* tmp_data, double* u_data, double *Vext_data, double *A_diag_data,int* A_diag_j,int* A_offd_data,int* A_diag_i, int* A_offd_i, int* A_offd_j, double* f_data,int n, int num_threads )
{

	int i,ii,j,jj,ns,ne,res,rest,size;
	j = blockIdx.x * blockDim.x + threadIdx.x;
		
	    size = n/num_threads;
	    rest = n - size*num_threads;
	    if (j < rest)
	    {
	       ns = j*size+j;
	       ne = (j+1)*size+j+1;
	    }
	    else
	    {
	       ns = j*size+rest;
	       ne = (j+1)*size+rest;
	    }

            for (i = ns; i < ne; i++)	/* interior points first */
            {

               /*-----------------------------------------------------------
                * If diagonal is nonzero, relax point i; otherwise, skip it.
                *-----------------------------------------------------------*/

               if ( A_diag_data[A_diag_i[i]] != 0)
               {
                  res = f_data[i];
                  for (jj = A_diag_i[i]+1; jj < A_diag_i[i+1]; jj++)
                  {
                     ii = A_diag_j[jj];
		     if (ii >= ns && ii < ne)
		     {
                        res -= A_diag_data[jj] * u_data[ii];
		     }
		     else
                        res -= A_diag_data[jj] * tmp_data[ii];
                  }
                  for (jj = A_offd_i[i]; jj < A_offd_i[i+1]; jj++)
                  {
                     ii = A_offd_j[jj];
                     res -= A_offd_data[jj] * Vext_data[ii];
                  }
                  u_data[i] = res / A_diag_data[A_diag_i[i]];
               }
            }
            for (i = ne-1; i > ns-1; i--)	/* interior points first */
            {

               /*-----------------------------------------------------------
                * If diagonal is nonzero, relax point i; otherwise, skip it.
                *-----------------------------------------------------------*/

               if ( A_diag_data[A_diag_i[i]] != 0)
               {
                  res = f_data[i];
                  for (jj = A_diag_i[i]+1; jj < A_diag_i[i+1]; jj++)
                  {
                     ii = A_diag_j[jj];
		     if (ii >= ns && ii < ne)
		     {
                        res -= A_diag_data[jj] * u_data[ii];
		     }
		     else
                        res -= A_diag_data[jj] * tmp_data[ii];
                  }
                  for (jj = A_offd_i[i]; jj < A_offd_i[i+1]; jj++)
                  {
                     ii = A_offd_j[jj];
                     res -= A_offd_data[jj] * Vext_data[ii];
                  }
                  u_data[i] = res / A_diag_data[A_diag_i[i]];
               }
            }
             
  }    


extern "C"
{

void runKernel(double* tmp_data, double* u_data, double *Vext_data, double *A_diag_data,int* A_diag_j,double* A_offd_data,int* A_diag_i, int* A_offd_i, int* A_offd_j, double* f_data,int num_cols_offd,int n, int num_threads)
{

     double *ctmp_data, *cu_data, *cVext_data, *cA_diag_data;
           int *cA_diag_j, *cA_offd_data,*cA_diag_i, *cA_offd_i, *cA_offd_j; 
	   double* cf_data;

	  int size = n/num_threads;
	  int rest = n - size*num_threads;
          int ne;
     

       
    if ((num_threads-1) < rest)
     {
      
        ne = num_threads*size+num_threads;
     }
     else
     {
       
        ne = (num_threads)*size+rest;
     }      
         
         

           cudaMalloc((void**)&ctmp_data,n*sizeof(double));
           cudaMalloc((void**)&cu_data,n*sizeof(double));
           cudaMalloc((void**)&cVext_data,sizeof(double)*num_cols_offd);
           cudaMalloc((void**)&cA_diag_data,(A_diag_i[ne])*sizeof(double));
           cudaMalloc((void**)&cA_diag_i,(ne)*sizeof(int));
	   cudaMalloc((void**)&cA_diag_j,(A_diag_i[ne])*sizeof(int));
           cudaMalloc((void**)&cA_offd_i,(ne)*sizeof(int));
           cudaMalloc((void**)&cA_offd_data,(A_offd_i[ne]+1)*sizeof(int));
           cudaMalloc((void**)&cA_offd_j,(A_offd_i[ne])*sizeof(int));
           cudaMalloc((void**)&cf_data,(ne)*sizeof(double));

	   cudaMemcpy(ctmp_data,tmp_data,n*sizeof(double),cudaMemcpyHostToDevice);
           cudaMemcpy(cu_data,u_data,n*sizeof(double),cudaMemcpyHostToDevice);
           cudaMemcpy(cA_diag_i,A_diag_i,ne*sizeof(int),cudaMemcpyHostToDevice);
            cudaMemcpy(cA_offd_i,A_offd_i,ne*sizeof(int),cudaMemcpyHostToDevice);
            cudaMemcpy(cf_data,f_data,ne*sizeof(double),cudaMemcpyHostToDevice);  
           cudaMemcpy(cA_diag_j,A_diag_j,A_diag_i[ne]*sizeof(int),cudaMemcpyHostToDevice); 
           cudaMemcpy(cA_offd_j,A_offd_j,A_offd_i[ne]*sizeof(int),cudaMemcpyHostToDevice);
            cudaMemcpy(cA_diag_data,A_diag_data,A_diag_i[ne]*sizeof(double),cudaMemcpyHostToDevice);  
            cudaMemcpy(cA_offd_data,A_offd_data,A_offd_i[ne]*sizeof(int),cudaMemcpyHostToDevice); 
	    cudaMemcpy(cVext_data,Vext_data,num_cols_offd*sizeof(double),cudaMemcpyHostToDevice);


	    relax_on_CUDA <<<1,num_threads>>> (ctmp_data, cu_data, cVext_data, cA_diag_data,cA_diag_j,cA_offd_data,cA_diag_i,cA_offd_i,cA_offd_j, cf_data, n,num_threads);
    
            cudaMemcpy(u_data,cu_data,n*sizeof(double),cudaMemcpyDeviceToHost);
           
            cudaFree(ctmp_data);
            cudaFree(cu_data);
	    cudaFree(cVext_data);
	    cudaFree(cA_diag_data);
	    cudaFree(cA_diag_i);
	    cudaFree(cA_diag_j);
	    cudaFree(cA_offd_i);
	    cudaFree(cA_offd_data);
	    cudaFree(cA_offd_j);
	    cudaFree(cf_data);

          


           return;


}



}