#include <stdio.h>
#include "cuda.h"
#include "cuda_runtime_api.h" 


__global__ void vecturreductioncuda(double* x_data,double* y_data,double* result,int size)
{

  int j,i;
  
  j = blockIdx.x * blockDim.x + threadIdx.x;

  result[j] =0;

  for(i = j*size ; i < ((j+1)*size); i++)
    {
      result[j] += x_data[i] * y_data[i];

    }


}

extern "C"
{
double vectorreductioncalloncuda(double* x_data, double* y_data, int size)
{

  double *cx_data, *cy_data,*cresult,finalresult,*result; 
  int i,num_threads = 512;

  finalresult = 0.0;
    int division = (size/num_threads);

   if(division == 0)
  {
	 for(i=(division*num_threads); i < size; i++)
	{
	  finalresult += x_data[i] * y_data[i];
	}	
  

   }
  else
  {

  result = (double*)malloc(sizeof(double)*num_threads);
  cudaMalloc((void**)&cx_data,size*sizeof(double));
  cudaMalloc((void**)&cy_data,size*sizeof(double));
  cudaMemcpy(cx_data,x_data,size*sizeof(double),cudaMemcpyHostToDevice);
  cudaMemcpy(cy_data,y_data,size*sizeof(double),cudaMemcpyHostToDevice);
  cudaMalloc((void**)&cresult,num_threads*sizeof(double));
  

  vecturreductioncuda <<<1,num_threads>>>(cx_data,cy_data,cresult,division);

  cudaMemcpy(result,cresult,num_threads*sizeof(double),cudaMemcpyDeviceToHost);
    


  for(i=0; i<num_threads; i++)
    {

      finalresult += result[i];
    } 
  



  if(size%num_threads != 0)
    {
      for(i=(division*num_threads); i < size; i++)
	{
	  finalresult += x_data[i] * y_data[i];
	} 

    }

  cudaFree(cx_data);
  cudaFree(cy_data);
  cudaFree(cresult);
  free(result);
  } 

  return finalresult;

}
}


__global__ void multiplyarrayoncuda(double* x_data, double* y_data, double alpha, int size)
{
   int j,i;
  
  j = blockIdx.x * blockDim.x + threadIdx.x;



  for(i = j*size ; i < ((j+1)*size); i++)
    {
     
      y_data[i] += alpha*x_data[i];

    }



}


extern "C"
{
void multiplearrayvector(double* x_data, double* y_data, double alpha, int size)
{
    double *cx_data, *cy_data;
     int i;
    int num_threads = 512;
     int division = (size/num_threads);

   if(division == 0)
  {
	for(i=(division*num_threads); i < size; i++)
	{
	  y_data[i]  += alpha * x_data[i];
	} 	

  }
  else
  {
 cudaMalloc((void**)&cx_data,size*sizeof(double));
     cudaMalloc((void**)&cy_data,size*sizeof(double));
  cudaMemcpy(cx_data,x_data,size*sizeof(double),cudaMemcpyHostToDevice);
  cudaMemcpy(cy_data,y_data,size*sizeof(double),cudaMemcpyHostToDevice);

      
  multiplyarrayoncuda <<<1, num_threads>>>(cx_data,cy_data,alpha,division);

   cudaMemcpy(y_data,cy_data,sizeof(double)*size,cudaMemcpyDeviceToHost); 


      if(size%num_threads != 0)
      {
	for(i=(division*num_threads); i < size; i++)
	{
	  y_data[i]  += alpha * x_data[i];
	} 

    }

  cudaFree(cx_data);
  cudaFree(cy_data);

  }
  
   return; 

 } 

    

}

__global__ void genericmutliply(double* bin, int size, double multiplicand)
{
   int i,j; 

   j = blockIdx.x * blockDim.x + threadIdx.x;
   
   for(i=j*size; i < (j+1)*size; i++)
   {
	bin[i] *= multiplicand;

   }
    

}

extern "C"
{
void generalmultiply(double* bin, int range, double multiplicand)
{
   
    double* cbin;
    int num_threads;
    int i=0;
   
     num_threads = 512;  

     

    int size = range/num_threads;


    if(size == 0)
    {
       for(i=size*num_threads; i< range; i++)
	{
	    bin[i] *= multiplicand;

	}


    }
    else
    {
 
    

    cudaMalloc((void**)&cbin,sizeof(double)*range);

    cudaMemcpy(cbin,bin,sizeof(double)*range,cudaMemcpyHostToDevice);
    
    genericmutliply<<<1,num_threads>>>(cbin,size,multiplicand);

    cudaMemcpy(bin,cbin,sizeof(double)*range,cudaMemcpyDeviceToHost);
    

    if(range%num_threads != 0)
   {
	for(i=size*num_threads; i< range; i++)
	{
	    bin[i] *= multiplicand;

	}
   }
   cudaFree(cbin);
 

  }


}

}

__global__ void genericcomparator(double* bin, int* comparator, int comparand, double multiplicand, int size)
{
 int i,j;
  j = blockIdx.x * blockDim.x + threadIdx.x;

  for(i=j*size; i < (j+1)*size; i++)
  {
	if(comparator[i] == comparand)
	{
	   bin[i] *= multiplicand;
	
	}
	
  }


}

extern "C"
{
void generalcomparatormultiply(double* bin, int* comparatorfromhost, int comparand, double multiplicand, int range)
{
   
    double* cbin;
    int num_threads;
    int i=0;
    int* comparator;
   
     num_threads = 512;  

     

    int size = range/num_threads;

  printf("%d\n",size);

    if(size == 0)
    {
       for(i=size*num_threads; i< range; i++)
	{
	   if(comparatorfromhost[i] == comparand)
	   {
		bin[i] *= multiplicand;
	   }	

	}


    }
    else
    {
    cudaMalloc((void**)&comparator,sizeof(int)*range);

    cudaMalloc((void**)&cbin,sizeof(double)*range);

    cudaMemcpy(comparator,comparatorfromhost,sizeof(int)*range,cudaMemcpyHostToDevice);

    cudaMemcpy(cbin,bin,sizeof(double)*range,cudaMemcpyHostToDevice);

   
    
    genericcomparator<<<1,num_threads>>>(cbin, comparator, comparand, multiplicand, size);

    cudaMemcpy(bin,cbin,sizeof(double)*range,cudaMemcpyDeviceToHost);
    
    

    if(range%num_threads != 0)
   {
	for(i=size*num_threads; i< range; i++)
	{
	  if(comparatorfromhost[i] == comparand)
	   {
	    
	    bin[i] *= multiplicand;

	    }

	}
   } 

   cudaFree(comparator);
 cudaFree(cbin);

  }

 
 



}

}

